@@ -34,16 +34,20 @@ from typing import Dict
 from typing import List
 from typing import Optional
 from typing import overload
+from typing import Tuple
 from typing import Type
 from typing import Union
 from uuid import uuid4
 
-import pytest
+import deprecation
 import testinfra
-from _pytest.mark.structures import MarkDecorator
-from _pytest.mark.structures import ParameterSet
+from _pytest.mark import Mark
+from _pytest.mark import MarkDecorator
+from _pytest.mark import ParameterSet
+from filelock import BaseFileLock
 from filelock import FileLock
-
+from pytest import param
+from pytest_container.helpers import get_always_pull_option
 from pytest_container.inspect import ContainerHealth
 from pytest_container.inspect import ContainerInspect
 from pytest_container.inspect import PortForwarding
@@ -52,6 +56,13 @@ from pytest_container.logging import _logger
 from pytest_container.runtime import get_selected_runtime
 from pytest_container.runtime import OciRuntimeBase
 
+if sys.version_info >= (3, 8):
+    from importlib import metadata
+    from typing import Literal
+else:
+    import importlib_metadata as metadata
+    from typing_extensions import Literal
+
 
 @enum.unique
 class ImageFormat(enum.Enum):
@@ -68,6 +79,11 @@ class ImageFormat(enum.Enum):
         return "oci" if self == ImageFormat.OCIv1 else "docker"
 
 
+def lock_host_port_search(rootdir: Path) -> BaseFileLock:
+    """Generate a filelock for finding free ports on the host."""
+    return FileLock(rootdir / "port_check.lock")
+
+
 def create_host_port_port_forward(
     port_forwards: List[PortForwarding],
 ) -> List[PortForwarding]:
@@ -79,33 +95,34 @@ def create_host_port_port_forward(
     """
     finished_forwards: List[PortForwarding] = []
 
-    # list of sockets that will be created and cleaned up afterwards
-    # We have to defer the cleanup, as otherwise the OS might give us a
-    # previously freed socket again. But it will not do that, if we are still
-    # listening on it.
-    sockets: List[socket.socket] = []
+    # We have to defer the cleanup of all sockets via an ExitStack, as otherwise
+    # the OS might give us a previously freed port again. But it will not do
+    # that, if we are still listening on it
+    with contextlib.ExitStack() as stack:
+        for port in port_forwards:
+            if socket.has_ipv6 and (":" in port.bind_ip or not port.bind_ip):
+                family = socket.AF_INET6
+            else:
+                family = socket.AF_INET
 
-    for port in port_forwards:
-
-        sock = socket.socket(
-            family=socket.AF_INET6 if socket.has_ipv6 else socket.AF_INET,
-            type=port.protocol.SOCK_CONST,
-        )
-        sock.bind(("", 0))
-
-        port_num: int = sock.getsockname()[1]
-
-        finished_forwards.append(
-            PortForwarding(
-                container_port=port.container_port,
-                protocol=port.protocol,
-                host_port=port_num,
+            sock = stack.enter_context(
+                socket.socket(
+                    family=family,
+                    type=port.protocol.SOCK_CONST,
+                )
             )
-        )
-        sockets.append(sock)
+            sock.bind((port.bind_ip, max(0, port.host_port)))
 
-    for sock in sockets:
-        sock.close()
+            port_num: int = sock.getsockname()[1]
+
+            finished_forwards.append(
+                PortForwarding(
+                    container_port=port.container_port,
+                    protocol=port.protocol,
+                    host_port=port_num,
+                    bind_ip=port.bind_ip,
+                )
+            )
 
     assert len(port_forwards) == len(finished_forwards)
     return finished_forwards
@@ -440,10 +457,21 @@ class ContainerBase:
     custom_entry_point: Optional[str] = None
 
     #: List of additional flags that will be inserted after
-    #: `docker/podman run -d`. The list must be properly escaped, e.g. as
-    #: created by `shlex.split`
+    #: `docker/podman run -d` and before the image name (i.e. these arguments
+    #: are not passed to the entrypoint or ``CMD``). The list must be properly
+    #: escaped, e.g. as created by ``shlex.split``.
     extra_launch_args: List[str] = field(default_factory=list)
 
+    #: List of additional arguments that are passed to the ``CMD`` or
+    #: entrypoint. These arguments are inserted after the :command:`docker/podman
+    #: run -d $image` on launching the image.
+    #: The list must be properly escaped, e.g. by passing the string through
+    #: ``shlex.split``.
+    #: The arguments must not cause the container to exit early. It must remain
+    #: active in the background, otherwise this library will not function
+    #: properly.
+    extra_entrypoint_args: List[str] = field(default_factory=list)
+
     #: Time for the container to become healthy (the timeout is ignored
     #: when the container image defines no ``HEALTHCHECK`` or when the timeout
     #: is below zero).
@@ -564,6 +592,8 @@ class ContainerBase:
         else:  # pragma: no cover
             assert False, "This branch must be unreachable"  # pragma: no cover
 
+        cmd.extend(self.extra_entrypoint_args)
+
         return cmd
 
     @property
@@ -600,12 +630,15 @@ class ContainerBaseABC(ABC):
 
     @abstractmethod
     def prepare_container(
-        self, rootdir: Path, extra_build_args: Optional[List[str]]
+        self,
+        container_runtime: OciRuntimeBase,
+        rootdir: Path,
+        extra_build_args: Optional[List[str]],
     ) -> None:
         """Prepares the container so that it can be launched."""
 
     @abstractmethod
-    def get_base(self) -> "Container":
+    def get_base(self) -> "Union[Container, DerivedContainer]":
         """Returns the Base of this Container Image. If the container has no
         base, then ``self`` is returned.
 
@@ -624,19 +657,30 @@ class ContainerBaseABC(ABC):
 class Container(ContainerBase, ContainerBaseABC):
     """This class stores information about the Container Image under test."""
 
-    def pull_container(self) -> None:
+    def pull_container(self, container_runtime: OciRuntimeBase) -> None:
         """Pulls the container with the given url using the currently selected
         container runtime"""
-        runtime = get_selected_runtime()
-        _logger.debug("Pulling %s via %s", self.url, runtime.runner_binary)
-        check_output([runtime.runner_binary, "pull", self.url])
+        _logger.debug(
+            "Pulling %s via %s", self.url, container_runtime.runner_binary
+        )
+        check_output([container_runtime.runner_binary, "pull", self.url])
 
     def prepare_container(
-        self, rootdir: Path, extra_build_args: Optional[List[str]] = None
+        self,
+        container_runtime: OciRuntimeBase,
+        rootdir: Path,
+        extra_build_args: Optional[List[str]] = None,
     ) -> None:
         """Prepares the container so that it can be launched."""
-        if not self._is_local:
-            self.pull_container()
+        if self._is_local:
+            return
+
+        if get_always_pull_option():
+            self.pull_container(container_runtime)
+            return
+
+        if call([container_runtime.runner_binary, "inspect", self.url]) != 0:
+            self.pull_container(container_runtime)
 
     def get_base(self) -> "Container":
         return self
@@ -691,28 +735,29 @@ class DerivedContainer(ContainerBase, ContainerBaseABC):
             or f"container derived from {self.base.__str__()}"
         )
 
-    def get_base(self) -> Container:
-        """Return the recursive base of this derived container (i.e. if the base
-        if this container is a derived one, then it takes the base of the
-        base).
-
-        """
+    def get_base(self) -> Union[Container, "DerivedContainer"]:
+        """Return the base of this derived container."""
         if isinstance(self.base, str):
             return Container(url=self.base)
-        return self.base.get_base()
+        return self.base
 
     def prepare_container(
-        self, rootdir: Path, extra_build_args: Optional[List[str]] = None
+        self,
+        container_runtime: OciRuntimeBase,
+        rootdir: Path,
+        extra_build_args: Optional[List[str]] = None,
     ) -> None:
         _logger.debug("Preparing derived container based on %s", self.base)
         if isinstance(self.base, str):
             # we need to pull the container so that the inspect in the launcher
             # doesn't fail
             Container(url=self.base).prepare_container(
-                rootdir, extra_build_args
+                container_runtime, rootdir, extra_build_args
             )
         else:
-            self.base.prepare_container(rootdir, extra_build_args)
+            self.base.prepare_container(
+                container_runtime, rootdir, extra_build_args
+            )
 
         runtime = get_selected_runtime()
 
@@ -720,6 +765,9 @@ class DerivedContainer(ContainerBase, ContainerBaseABC):
         # tags are added
         if not self.containerfile and not self.add_build_tags:
             base = self.get_base()
+            base.prepare_container(
+                container_runtime, rootdir, extra_build_args
+            )
             self.container_id, self.url = base.container_id, base.url
             return
 
@@ -744,14 +792,10 @@ class DerivedContainer(ContainerBase, ContainerBaseABC):
                 containerfile.write(containerfile_contents)
 
             cmd = runtime.build_command
-            if "buildah" in runtime.build_command:
+            if "podman" in runtime.runner_binary:
                 if self.image_format is not None:
                     cmd += ["--format", str(self.image_format)]
                 else:
-                    assert (
-                        "podman" in runtime.runner_binary
-                    ), "The runner for buildah should be podman"
-
                     if not runtime.supports_healthcheck_inherit_from_base:
                         warnings.warn(
                             UserWarning(
@@ -867,9 +911,67 @@ def container_to_pytest_param(
     :py:attr:`~pytest_container.container.ContainerBase.container_id`)
 
     """
-    return pytest.param(container, marks=marks or [], id=str(container))
+    return param(container, marks=marks or [], id=str(container))
 
 
+@overload
+def container_and_marks_from_pytest_param(
+    ctr_or_param: Container,
+) -> Tuple[Container, Literal[None]]:
+    ...
+
+
+@overload
+def container_and_marks_from_pytest_param(
+    ctr_or_param: DerivedContainer,
+) -> Tuple[DerivedContainer, Literal[None]]:
+    ...
+
+
+@overload
+def container_and_marks_from_pytest_param(
+    ctr_or_param: ParameterSet,
+) -> Tuple[
+    Union[Container, DerivedContainer],
+    Optional[Collection[Union[MarkDecorator, Mark]]],
+]:
+    ...
+
+
+def container_and_marks_from_pytest_param(
+    ctr_or_param: Union[ParameterSet, Container, DerivedContainer],
+) -> Tuple[
+    Union[Container, DerivedContainer],
+    Optional[Collection[Union[MarkDecorator, Mark]]],
+]:
+    """Extracts the :py:class:`~pytest_container.container.Container` or
+    :py:class:`~pytest_container.container.DerivedContainer` and the
+    corresponding marks from a `pytest.param
+    <https://docs.pytest.org/en/stable/reference.html?#pytest.param>`_ and
+    returns both.
+
+    If ``param`` is either a :py:class:`~pytest_container.container.Container`
+    or a :py:class:`~pytest_container.container.DerivedContainer`, then param is
+    returned directly and the second return value is ``None``.
+
+    """
+    if isinstance(ctr_or_param, (Container, DerivedContainer)):
+        return ctr_or_param, None
+
+    if len(ctr_or_param.values) > 0 and isinstance(
+        ctr_or_param.values[0], (Container, DerivedContainer)
+    ):
+        return ctr_or_param.values[0], ctr_or_param.marks
+
+    raise ValueError(f"Invalid pytest.param values: {ctr_or_param.values}")
+
+
+@deprecation.deprecated(
+    deprecated_in="0.4.0",
+    removed_in="0.5.0",
+    current_version=metadata.version("pytest_container"),
+    details="use container_and_marks_from_pytest_param instead",
+)  # type: ignore
 def container_from_pytest_param(
     param: Union[ParameterSet, Container, DerivedContainer],
 ) -> Union[Container, DerivedContainer]:
@@ -960,7 +1062,7 @@ class ContainerLauncher:
         try:
             lock.acquire()
             self.container.prepare_container(
-                self.rootdir, self.extra_build_args
+                self.container_runtime, self.rootdir, self.extra_build_args
             )
         except:
             release_lock()
@@ -991,7 +1093,7 @@ class ContainerLauncher:
         # port forwards must be launched while the lock is being held. Otherwise
         # another container could pick the same ports before this one launches.
         if forwarded_ports and self._expose_ports:
-            with FileLock(self.rootdir / "port_check.lock"):
+            with lock_host_port_search(self.rootdir):
                 self._new_port_forwards = create_host_port_port_forward(
                     forwarded_ports
                 )
