@@ -19,10 +19,9 @@ from typing import Optional
 from typing import TYPE_CHECKING
 from typing import Union
 
-import pytest
 import testinfra
 from _pytest.mark.structures import ParameterSet
-
+from pytest import param
 from pytest_container.inspect import BindMount
 from pytest_container.inspect import Config
 from pytest_container.inspect import ContainerHealth
@@ -45,7 +44,6 @@ else:
 # cached_property from stdlib and we'll be fine
 if TYPE_CHECKING:  # pragma: no cover
     from functools import cached_property
-    from .container import ContainerBase
 else:
     try:
         from functools import cached_property
@@ -67,7 +65,7 @@ class ToParamMixin:
 
     def to_pytest_param(self) -> ParameterSet:
         """Convert this class into a ``pytest.param``"""
-        return pytest.param(self, id=str(self), marks=self.marks or ())
+        return param(self, id=str(self), marks=self.marks or ())
 
 
 @dataclass(frozen=True)
@@ -387,7 +385,11 @@ class OciRuntimeBase(_OciRuntimeBase, OciRuntimeABC, ToParamMixin):
                         host_port=int(bindings[0]["HostPort"]),
                     )
                 )
-        return ContainerNetworkSettings(ports=ports)
+
+        net_settings = container_inspect["NetworkSettings"]
+        ip = net_settings.get("IPAddress") or None
+
+        return ContainerNetworkSettings(ports=ports, ip_address=ip)
 
     @staticmethod
     def _mounts_from_inspect(
@@ -448,27 +450,27 @@ class PodmanRuntime(OciRuntimeBase):
 
     """
 
-    _runtime_functional = (
-        LOCALHOST.run("podman ps").succeeded
-        and LOCALHOST.run("buildah").succeeded
-    )
+    _runtime_functional = LOCALHOST.run("podman ps").succeeded
+    _buildah_functional = LOCALHOST.run("buildah").succeeded
 
     @staticmethod
     def _runtime_error_message() -> str:
         if PodmanRuntime._runtime_functional:
             return ""
+
         podman_ps = LOCALHOST.run("podman ps")
-        if not podman_ps.succeeded:
-            return str(podman_ps.stderr)
-        buildah = LOCALHOST.run("buildah")
         assert (
-            not buildah.succeeded
-        ), "buildah command must not succeed as PodmanRuntime is not functional"
-        return str(buildah.stderr)
+            not podman_ps.succeeded
+        ), "podman runtime is not functional, but 'podman ps' succeeded"
+        return str(podman_ps.stderr)
 
     def __init__(self) -> None:
         super().__init__(
-            build_command=["buildah", "bud", "--layers", "--force-rm"],
+            build_command=(
+                ["buildah", "bud", "--layers", "--force-rm"]
+                if self._buildah_functional
+                else ["podman", "build", "--layers", "--force-rm"]
+            ),
             runner_binary="podman",
             _runtime_functional=self._runtime_functional,
         )
@@ -487,29 +489,41 @@ class PodmanRuntime(OciRuntimeBase):
         #   https://github.com/containers/buildah/blob/main/CHANGELOG.md#v1250-2022-03-25
         # - podman 4.1.0 bundles buildah >= 1.25.0
         #   https://github.com/containers/podman/blob/main/RELEASE_NOTES.md#misc-8
-        return self.version >= Version(
-            4, 1, 0
-        ) and _get_buildah_version() >= Version(1, 25, 0)
+        podman_recent_enough = self.version >= Version(4, 1, 0)
+
+        # if buildah isn't installed, don't check the buildah version
+        if not self._buildah_functional:
+            return podman_recent_enough
+
+        return podman_recent_enough and _get_buildah_version() >= Version(
+            1, 25, 0
+        )
 
     def inspect_container(self, container_id: str) -> ContainerInspect:
         inspect = self._get_container_inspect(container_id)
 
-        Conf = inspect["Config"]
+        config = inspect["Config"]
         healthcheck = None
-        if "Healthcheck" in Conf:
+        if "Healthcheck" in config:
             healthcheck = HealthCheck.from_container_inspect(
-                Conf["Healthcheck"]
+                config["Healthcheck"]
             )
 
+        entrypoint = config.get("Entrypoint")
+        if isinstance(entrypoint, str):
+            entrypoint = entrypoint.split()
+        if not entrypoint:
+            entrypoint = []
+
         conf = Config(
-            user=Conf["User"],
-            tty=Conf["Tty"],
-            cmd=Conf["Cmd"],
-            image=Conf["Image"],
-            entrypoint=Conf["Entrypoint"].split(),
-            labels=Conf["Labels"],
-            env=dict([env.split("=", maxsplit=1) for env in Conf["Env"]]),
-            stop_signal=self._stop_signal_from_inspect_conf(Conf),
+            user=config["User"],
+            tty=config["Tty"],
+            cmd=config["Cmd"],
+            image=config["Image"],
+            entrypoint=entrypoint,
+            labels=config["Labels"],
+            env=dict([env.split("=", maxsplit=1) for env in config["Env"]]),
+            stop_signal=self._stop_signal_from_inspect_conf(config),
             healthcheck=healthcheck,
         )
 
@@ -518,6 +532,7 @@ class PodmanRuntime(OciRuntimeBase):
         return ContainerInspect(
             config=conf,
             state=state,
+            name=inspect["Name"],
             id=inspect["Id"],
             path=inspect["Path"],
             args=inspect["Args"],
@@ -573,25 +588,25 @@ class DockerRuntime(OciRuntimeBase):
     def inspect_container(self, container_id: str) -> ContainerInspect:
         inspect = self._get_container_inspect(container_id)
 
-        Conf = inspect["Config"]
-        if Conf.get("Env"):
-            env = dict([env.split("=", maxsplit=1) for env in Conf["Env"]])
+        config = inspect["Config"]
+        if config.get("Env"):
+            env = dict([env.split("=", maxsplit=1) for env in config["Env"]])
         else:
             env = {}
         healthcheck = None
-        if "Healthcheck" in Conf:
+        if "Healthcheck" in config:
             healthcheck = HealthCheck.from_container_inspect(
-                Conf["Healthcheck"]
+                config["Healthcheck"]
             )
 
         conf = Config(
-            user=Conf["User"],
-            tty=Conf["Tty"],
-            cmd=Conf["Cmd"],
-            image=Conf["Image"],
-            entrypoint=Conf["Entrypoint"],
-            labels=Conf["Labels"],
-            stop_signal=self._stop_signal_from_inspect_conf(Conf),
+            user=config["User"],
+            tty=config["Tty"],
+            cmd=config["Cmd"],
+            image=config["Image"],
+            entrypoint=config["Entrypoint"],
+            labels=config["Labels"],
+            stop_signal=self._stop_signal_from_inspect_conf(config),
             env=env,
             healthcheck=healthcheck,
         )
@@ -601,6 +616,8 @@ class DockerRuntime(OciRuntimeBase):
         return ContainerInspect(
             config=conf,
             state=state,
+            # docker prefixes the name with a / for reasons…
+            name=inspect["Name"].lstrip("/"),
             id=inspect["Id"],
             path=inspect["Path"],
             args=inspect["Args"],
@@ -619,7 +636,7 @@ def get_selected_runtime() -> OciRuntimeBase:
 
     If neither docker nor podman are available, then a ValueError is raised.
     """
-    podman_exists = LOCALHOST.exists("podman") and LOCALHOST.exists("buildah")
+    podman_exists = LOCALHOST.exists("podman")
     docker_exists = LOCALHOST.exists("docker")
 
     runtime_choice = getenv("CONTAINER_RUNTIME", "podman").lower()
